# Copyright (c) ONNX Project Contributors
#
# SPDX-License-Identifier: Apache-2.0

# This file needs to be here to enable mypy type checking for this folder.
# It doesn't make this a python module for installation purposes as a __init__.pyi would.
